#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pr/pfLinMath.h>
#include <Performer/pr/pfType.h>
#include "pfcave.h"
#include "NavPath.hh"

const	float	SPEED = 2.0f;
const	float	rotSensitivity =  0.2f;

void NavPath::navigate(pfDCS *dcs)
{
	float	jx=CAVE_JOYSTICK_X, jy=CAVE_JOYSTICK_Y;
	float	qx, qy;
	//int	abutton = 0;
	//static	int	apressed = 0;
	//static	int	apressedOld = 0;

/*
	// test button 1 (LEFT) for press.
	abutton = CAVEBUTTON1;
	if ( abutton ) {
		apressed = 1;
	}
	else {
		apressed = 0;
	}
	// check for change of state.
	if ( apressed != apressedOld ) {
		fprintf(stderr,"Changed state... (%d)\n", apressed);
		apressedOld = apressed;
		if ( apressed ) {
			dropOutDCS();	// save to the file...
		}
	}
*/
	// if no buttons pressed, then do the nav thang
	if ( !CAVEBUTTON1 && !CAVEBUTTON2 && !CAVEBUTTON3 ) {
		if ( PF_ABS(jx) > 0.1) {
			CAVENavRot(rotSensitivity*jx,'z');
		}
		if ( PF_ABS(jy) > 0.1) {
			float w[3];
			CAVEGetVector(CAVE_WAND_FRONT,w);
			jx *= 1.1;
			jy *= 1.1;
			qx = jx * PF_ABS(jx);
			qy = jy * PF_ABS(jy);
			CAVENavTranslate( w[0]*qx*SPEED, w[1]*qy*SPEED,
							 w[2]*qy*SPEED);
		}
	}

	// if cavebutton2, then rotate.. sort of like a grab
/*
	if ( CAVEBUTTON2 ) {
		float w[3];
		CAVEGetVector(CAVE_WAND_FRONT,w);
		CAVENavRot(jy*rotSensitivity, 'x');
	}
*/
/*
	if (CAVEBUTTON3) {
		CAVENavLoadIdentity();
	}
*/

	if ( Shared->interp ) {
		fprintf(stderr,"JOY Y %12.5f\n", jy );
		interpDCS( jy );
	}
	pfCAVEDCSNavTransform(dcs); // ok, we'll skip using the base class to nav
}

/*********************************************************
name:    testDKey()
purpose: appears to test for press of "A" key
	if pressed, it calls "dropOutDCS() which
	writes a new DCS it the file "dcs.out"
**********************************************************/
void NavPath::testDKey(void)
{
	static	int	apressed = 0;
	static	int	apressedOld = 0;
	int	abutton = 0;

	// test AKEY.
	abutton = CAVEgetbutton(CAVE_AKEY);
	if ( abutton ) {
		apressed = 1;
	}
	else {
		apressed = 0;
	}
	// check for change of state.
	if ( apressed != apressedOld ) {
		fprintf(stderr,"Changed state... (%d)\n", apressed);
		apressedOld = apressed;
		if ( apressed ) {
			dropOutDCS();
		}
	}
}

// writes new DCS key to disk
void NavPath::dropOutDCS(void)
{
	static	int	first = 1;
	FILE	*fp;
	pfMatrix	mat;
	pfCoord	ic;
	pfDCS	*dcs;

	if ( first ) {
		fp = fopen("dcs.out","w");
		first = 0;
	}
	else {
		fp = fopen("dcs.out","a");
	}

	dcs = new pfDCS();

	pfCAVEDCSNavTransform(dcs);
	dcs->getMat( mat );
	mat.getOrthoCoord( &ic );
	printf("------------------------\n");
	printVec3( ic.hpr, "Interpolated HPR" );
	printVec3( ic.xyz, "Interpolated XYZ" );
	fprintf(fp,"%12.5f %12.5f %12.5f ", 
		ic.xyz[PF_X], ic.xyz[PF_Y], ic.xyz[PF_Z]);
	fprintf(fp,"%12.5f %12.5f %12.5f\n", 
		ic.hpr[PF_H], ic.hpr[PF_P], ic.hpr[PF_R]);
	fclose(fp);
}

// printVec3(); Utility to print out a pfVec3 w/label
// Mods: Just made it print to stderr... (unbuffered IO)
void NavPath::printVec3( const pfVec3 &v, char *label )
{
	fprintf(stderr,"%s   ", label);
	fprintf(stderr,"Vec3 :\n");
	fprintf(stderr,"    %12.5f %12.5f %12.5f\n", v[0], v[1], v[2]);
}


// Allocates memory for the "Shared" data structure
void NavPath::InitSharedMemory()
{
	void 	*arena;

	/* Get pointer to shared memory malloc arena created by pfInit()*/
	arena = pfGetSharedArena();
	/*
	 * Malloc structures now, before pfConfig(), 
	 * so forked processes have same pointers.
	*/
	Shared = (shared *) pfCalloc(1, sizeof(shared), arena);
}

// initializes the "Shared" structure
void NavPath::initShared(void)
{
	if (Shared != NULL) {
		Shared->numDCSMAX = 25; // start with 5
		Shared->interp = 0;
	}
}

// Creates a DCS
void NavPath::BuildDCS()
{
	int	i;
	int	num;
	pfCoord	c;
	FILE	*fp;
	float	x, y, z, h, p, r;

	// load DCS with values.
	fp = fopen("COORDS","r");
	if ( fp == NULL ) {
	  fprintf(stderr,"Error opening file\n");
	}
	else
	  fprintf(stderr,"Opened COORDS file\n");
	fscanf(fp,"%d", &num );
	// malloc from heap.
	Shared->numDCSMAX = num;

	Shared->DCSarray = (pfDCS **)pfMalloc( sizeof(pfDCS *) * Shared->numDCSMAX, 
									NULL );	
	for ( i = 0; i < Shared->numDCSMAX; i++ ) {
		Shared->DCSarray[i] = new pfDCS();
	}

	for ( i = 0; i < Shared->numDCSMAX; i++ ) {
		fscanf(fp,"%f %f %f %f %f %f", 
				&x, &y, &z, &h, &p, &r);
		printf("%8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
			x, y, z, h, p, r);
		c.xyz.set( x, y, z);
		c.hpr.set( h, p, r);
		Shared->DCSarray[i]->setCoord( &c );
		sprintf(buffer,"DCS %3d", i);
		xprintDCS( Shared->DCSarray[i], buffer);
	}
	fclose (fp); 
}

// Print out the contents of a DCS
void NavPath::xprintDCS( pfDCS *dcs, char *label )
{
	int	i;
	pfMatrix mat;

	dcs->getMat( mat );
	printf("------------\n");
	printf(" [%s] DCS Matrix (%p):\n", label, dcs);
	for ( i = 0; i < 4; i++ ) {
		printf("%12.5f %12.5f %12.5f %12.5f\n",
			mat[i][0], mat[i][1], mat[i][2], mat[i][3]);
	}
	printf("------------\n");
}


// Identify which process I am in...
void NavPath::WhichProcess( pid_t pid, char *routine, char *file, int line )
{

	if ( pid == Shared->appid ) {
		fprintf(stderr,
		"----->Application pid in %s [%s  %d] Parent %d\n", 
			routine, file, line, getppid());
	}
	else if ( pid == Shared->drawid ) {
		fprintf(stderr,
		"---->Drawing pid in %s [%s  %d] Parent %d\n", 
			routine, file, line,getppid());
	}
	else if ( pid == Shared->cullid ) {
		fprintf(stderr,
		"---->Culling pid in %s [%s  %d] Parent %d\n", 
			routine, file, line, getppid());
	}
	else {
		fprintf(stderr,
"--------->Unknowm process number (%d) %s [%s %d] Parent %d\n", 
			pid, routine, file, line, getppid());
	}
}

// Check for the "i" key
void NavPath::testIKey(void)
{
	static	int	apressed = 0;
	static	int	apressedOld = 0;
	int	abutton = 0;

	// test IKEY.
	abutton = CAVEgetbutton(CAVE_IKEY);
	if ( abutton ) {
		apressed = 1;
	}
	else {
		apressed = 0;
	}
	// check for change of state.
	if ( apressed != apressedOld ) {
		fprintf(stderr,"Changed state... (%d)\n", apressed);
		apressedOld = apressed;
		if ( apressed ) {
			Shared->interp = !Shared->interp;
		}
	}
}

// Interpolates between DCS's.  Position determined by param "t".
void NavPath::interpDCS( float t )
{

	static	float	cycle = 0.0f;
	static	float	delta = 0.01f;
	float	matrix[4][4];
	int	i;
	pfMatrix	mat;
	pfCoord		ic;

	if ( Shared->interp ) {
		cycle += delta * t;
		InterpolateCoord( Shared->DCSarray, cycle, &ic);
		mat.makeCoord( &ic );
	}
	for ( i = 0; i < 4; i++ ) {
		matrix[i][0] = mat[i][0];
		matrix[i][1] = mat[i][1];
		matrix[i][2] = mat[i][2];
		matrix[i][3] = mat[i][3];
	}
	CAVENavLoadMatrix( matrix );
}

// Interpolates between two coordinates in the array of DCS's.
void NavPath::InterpolateCoord( pfDCS **DCSarray, float t, pfCoord *c )
{
	int		left;
	int		right;
	float		fraction;
	pfMatrix	matLeft;
	pfMatrix	matRight;
	pfMatrix	mati;
	pfQuat		quatLeft;
	pfQuat		quatRight;
	pfQuat		iq;
	pfVec3		vLeft;
	pfVec3		vRight;
	pfVec3		v0;
	pfVec3		vi;
	pfCoord		ic;

	while ( t < 0.0 ) {
		t += Shared->numDCSMAX;
	}

	// put into the proper range.
	while ( t >= Shared->numDCSMAX ) {
		printf("before %12.5f  ", t);
		t -= Shared->numDCSMAX;
		printf("after %12.5f\n", t);
	}

	left = (int)t;
	if ( left >= Shared->numDCSMAX ) {
		return;
	}
	right = (left + 1) % Shared->numDCSMAX;
	fraction = t - (float)left;
	printf("------------------------------------------------------\n");
	printf(
	"Interpolation : %12.5f  Fraction %12.6f Left %3d  Right %3d\n",
		t, fraction, left, right);

	DCSarray[left]->getMat( matLeft );
	matLeft.getOrthoQuat( quatLeft );
	matLeft.getRow( 3, vLeft);

	DCSarray[right]->getMat( matRight );
	matRight.getOrthoQuat( quatRight );
	matRight.getRow( 3, vRight);

	// do the interpolation
	// the translations
	v0.sub( vRight, vLeft);	// v0 = v2 - v1;
	vi.addScaled(vLeft,fraction,v0);	// form v1 + fraction*v0
	iq.slerp(fraction, quatLeft, quatRight);
	mati.makeQuat( iq );
	mati.setRow(3, vi);
	mati.getOrthoCoord( &ic );
	c->xyz.copy( ic.xyz );
	c->hpr.copy( ic.hpr );
}

NavPath::NavPath(pfDCS *dcs,char *navfile):NavBase(dcs){
		InitSharedMemory();
		initShared();
		coordfile=NewString(navfile);
		fprintf(stderr,"Size of Shared %d\n", sizeof(shared));
		fprintf(stderr,"My pid is %d\n", getpid());
		Shared->appid = (int)pfGetPID(0, PFPROC_APP);
		fprintf(stderr,"Application pid %d\n", Shared->appid);
		Shared->drawid = (int)pfGetPID(0, PFPROC_DRAW);
		fprintf(stderr,"Draw        pid %d\n", Shared->drawid);
		Shared->cullid = (int)pfGetPID(0, PFPROC_CULL);
		fprintf(stderr,"Cull        pid %d\n", Shared->cullid);
		WhichProcess( getpid(), "Main", __FILE__, __LINE__);					     
		BuildDCS();
		Shared->interp=1; // set to interp mode
}

NavPath::~NavPath(){
	// need to delete every DCS in DCSarray 
	for(int i=0;i<Shared->numDCSMAX;i++){
		if((Shared->DCSarray[i])) 
			pfDelete(Shared->DCSarray[i]);
	}
	// now delete the pointer array
	pfDelete(Shared->DCSarray);
	// and the shared object (will ditch this object later)
	pfDelete(Shared);
	// and finally the name of the coordfile
	if(coordfile)
		delete coordfile;
}
	
void NavPath::update(){
		//testDKey(); // see if I should store a DCS
		//testIKey(); // see if I should switch to/from interpolated mode
		navigate(navDCS);
}
