//------Standard C/C++ libraries
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

//------John's Templates
#include "FlexArrayTmpl.H"
#include "pfFlexArrayTmpl.H"

//------Performer stuff
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfEarthSky.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfText.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfMaterial.h>
//#include <Performer/pr.h>
#include <Performer/pr/pfString.h>
#include <Performer/pr/pfFont.h>
//------Cave stuff
#include <pfcave.h>

pfGeoState *NoLightGState;
//------Utility subroutines for creating performer scene graph
pfDCS *NewScene(void);

// Typedefs
typedef pfDCS *pfDCSPtr;
struct sPoint { float x,y,z; };

union Point {
  float array[3];
  sPoint cart;
};

struct Atom {
  int type; // Al or Ni
  Point p;
};

class AtomList {
public:
  FlexArray<Atom> atoms;
  AtomList(){}
  // copy constructor (for copying Atom Lists)
  AtomList(AtomList &a){
    atoms = a.atoms;
  }
  // copy operator (for copying Atom Lists)
  AtomList &operator=(AtomList &a){
    if(&a!=this)
      atoms=a.atoms;
    return *this;
  }
  AtomList(char *filename){
    read(filename);
  }
  void read(char *filename){
    char buffer[1024];
    FILE *file=fopen(filename,"r");
    if(!file){
      printf("AtomList::read() could not open file %s\n",filename);
      return;
    }
    fgets(buffer,sizeof(buffer)-1,file);
    int len=atoi(buffer);
    fgets(buffer,sizeof(buffer)-1,file);
    while(fgets(buffer,sizeof(buffer)-1,file)){
      Atom atom;
      char name[32];
      sscanf(buffer,"%s %f %f %f",
	     name,&(atom.p.cart.x),&(atom.p.cart.y),&(atom.p.cart.z));
      if(strstr(name,"Al"))
	atom.type=1;
      else 
	atom.type=2;
      // printf("Appending %s\n",buffer); // debug
      atoms.append(atom);
    }
    fclose(file);
  }
  void print(){
    printf("Number of records = %u\n",atoms.getSize());
    for(int i=0;i<atoms.getSize();i++){
      printf("Type=%u: pos=%f:%f:%f\n",
	     atoms[i].type,
	     atoms[i].p.cart.x,atoms[i].p.cart.y,atoms[i].p.cart.z);
    }
  }
};

class TimeAtomList {
  FlexArray<AtomList> timesteps;
  static char *generateName(char *buffer,char *basename,int index){
    sprintf(buffer,"%s.%2u",basename,index);
    return buffer;
  }
public:
  TimeAtomList(char *namebase,int minindex,int maxindex){
    // declare the size of the timesteps array ahead of time
    timesteps.setSize(1+maxindex-minindex);
    printf("setting size to %u\n",1+maxindex-minindex);
    // Read in the data for the timesteps
    for(int idx=0,i=minindex;i<=maxindex;idx++,i++){
      char buffer[128];
      char *filename=generateName(buffer,namebase,i);
      printf("Loading file %s\n",filename);
      //(timesteps[idx]).print();
      (timesteps[idx]).read(filename);
      //(timesteps[idx]).print();
    }
  }
  // destructor (called when you "delete" this object)
  ~TimeAtomList(){ // nil for now
  }
  void getRange(float &min,float &max){
    min=0;
    max=(float)(timesteps.getSize()-1);
  }
  // linearly interpolate between timesteps
  // based on a floating-point "time" value
  // put the results in the array "atoms"
  void getAtoms(float time,FlexArray<Atom> &atoms){
    int timeindex=(int)time;
    if(timeindex==(timesteps.getSize()-1))
      timeindex--;
    float normdistance = time - (float)timeindex;
    // make sure input atom list is the same size
    // as the atom lists it is interpolating
    int natoms=(timesteps[timeindex]).atoms.getSize();
    if(atoms.getSize()!=natoms){
      atoms.setSize(natoms);
      for(int i=0;i<natoms;i++)
	atoms[i].type = timesteps[timeindex].atoms[i].type;
    }
    for(int i=0;i<natoms;i++){
      // just pointing to arrays that are deeply 
      // nested in data structures (just for
      // convenience)
      float *start=timesteps[timeindex].atoms[i].p.array;
      float *end=timesteps[timeindex+1].atoms[i].p.array;
      float *interp=atoms[i].p.array;
      // and here is the linear interpolation
      for(int j=0;j<3;j++)
	interp[j] = start[j] + (end[j]-start[j])*normdistance;
    }
  }
};

void print(FlexArray<Atom> &atoms){
  printf("Number of records = %u\n",atoms.getSize());
  for(int i=0;i<10;i++){
    printf("Type=%u: pos=%f:%f:%f\n",
	   atoms[i].type,
	   atoms[i].p.cart.x,atoms[i].p.cart.y,atoms[i].p.cart.z);
  }
}

int main(int argc,char *argv[]){ 
  //**************Local Vars
  pfDCS *movingdcs;
  pfDCS *stationarydcs;
  pfMatrix mat;
  
  // Label3D *navmode_label;
  pfSCS *labelscs;
  FlexArray<Atom> atomlist;

  TimeAtomList timeAtomList("fort",31,45);
  float min,max,val;
  timeAtomList.getRange(min,max);
  fprintf(stderr,"start interp\n");
  for(val=min;val<max;val+=0.1){
    timeAtomList.getAtoms(val,atomlist);
    //print(atomlist);
  }
  fprintf(stderr,"done\n");

  //***********Start the Cave Processes********************
  StartCave(&argc,argv,movingdcs,stationarydcs);
  //***********Setup Root-Level DCS's***********************
  mat.makeIdent();
  movingdcs->setMat(mat);
  stationarydcs->setMat(mat);
  mat.makeIdent();
  mat.preTrans(-4.0,5.0,8.0,mat);
  labelscs = new pfSCS(mat);
  stationarydcs->addChild(labelscs);
  // navmode_label = new Label3D(labelscs);
  // put the cone in the scene to mark center
  movingdcs->addChild(NewScene()); 
  while (!CAVEgetbutton(CAVE_ESCKEY)) {  
    static int nusers;
    pfSync();
    pfCAVEPreFrame();
    pfFrame();
    pfCAVEPostFrame();
    navigation.update();
    netHeadTrack->update();
    if(nusers!=*CAVENumUsers){
      nusers=*CAVENumUsers;
      cerr << "NumUsers=" << *CAVENumUsers << endl;
    }
  }
  StopCave();
  return 0;
}

void StartCave(int *argc,char *argv[],pfDCSPtr &movingdcs,pfDCSPtr &stationarydcs){
  //Local vars
  pfScene *scene = new pfScene;
  pfGeoState *gstate = new pfGeoState;
  pfLightSource *ls;
  pfSCS *lightSCS;
  pfMatrix mat;
  pfScene *scene;
  //***************Cave Startup
  pfInit();
  pfCAVEConfig(&argc,argv,NULL);
  CAVEFar = 1000;
  pfConfig();
  pfCAVEInitChannels();
  //**************Create the scene with a DCS at the top
  scenedcs = new pfDCS;
  cavedcs = new pfDCS;

  gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
  gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
  
  scene->setGState(gstate);
  mat.makeRot(45,1,0,0);
  scene->addChild(lightSCS=new pfSCS(mat));
  lightSCS->addChild(ls=new pfLightSource);
  ls->on();
  scene->addChild(scenedcs); 
  scene->addChild(cavedcs);
  chan->setScene(scene);

  pfCAVEMasterChan *chan=pfCAVEMasterChan();
  pfEarthSky *esky = new pfEarthSky();
  chan->setESky(esky);
  // GLOBAL!!!!
  NoLightGState = new pfGeoState;
  NoLightGState->setMode(PFSTATE_ENLIGHTING,PF_OFF);
}

void StopCave(){
  CAVEHalt();
  pfExit();
}

pfDCS *NewScene(){ // build the scene graph
  pfDCS *nav_dcs = new pfDCS;
  pfMaterial *material = new pfMaterial;
  pfGeoState *gstate = new pfGeoState;
  pfGeode *geom = new pfGeode;
  pfGeoSet *gset = pfdNewCone(64,pfGetSharedArena());
  pfVec4 *color = new pfVec4;

  material->setShininess(100.0);
  material->setColorMode(PFMTL_BOTH,PFMTL_CMODE_AMBIENT_AND_DIFFUSE);
  gstate->setAttr(PFSTATE_FRONTMTL,material);
  gstate->setAttr(PFSTATE_BACKMTL,material);
  color->set(0.0,0.2,0.8,1.0);
  gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,color,0);
  
  nav_dcs->addChild(geom);
  geom->addGSet(gset);
  gset->setGState(gstate);

  return nav_dcs;
}

/* spare junk */
{
  {
    float jx=CAVE_JOYSTICK_X;
    float jy=CAVE_JOYSTICK_Y;
    switch(mode){
    case MODE_SCALE_MODEL:{
      static float s=1.0;
      if(THRESH(jy,0.1)){
	if(jy>0)
	  s*= (1.0+jy)*SPEED;
	else 
	  s/= (1.0+jy)*SPEED;
	movingdcs->setScale(s,s,s);
      }
    }break;
    case MODE_SCALE_ATOMS:{
      static float sa=1.0;      
      if(THRESH(jy,0.1)){
	if(jy>0)
	  sa*= (1.0+jy)*SPEED;
	else 
	  sa/= (1.0+jy)*SPEED;
	atomdcs->setScale(sa,sa,sa);
      }
    }
    case MODE_NAVIGATE:{
      static double lasttime,thistime;
      static float rot;
      double dt;
      float w[3]; // wand vector
      lastTime=thisTime;
      thisTime=CAVEGetTime();
      dt=thisTime-lastTime;
      CaveWand.unitVectorDirection(w); // get the direction of the wand
      if(dt>0.5) dt=0; // no large jumps in time allowed!!
      if(THRESH(jy,0.1)){
	static float trans[3]={0.0,0.0,0.0};
	trans+=jy*30.0*dt;
	//navDCS->setRot(rot[0],rot[1],);
	navDCS->setTrans(trans*w[0],trans*w[1],trans*w[2]);
      }
      if(THRESH(jx,0.1)){
	static float rot=0.0;
	rot+=jx*30.0*dt; 
	movingdcs->setRot(0.0,0.0,rot);
	//navDCS->setTrans(trans[0],trans[1],trans[2]);
      }
    }
    }
  }
}
